/*		

	Foxtrot ?
 
	Variation autour d'un MMC ; 
	ergodique ;
	forward, backward, Baum-Welch avec r-chelonnement 
	des valeurs, (Rabiner 1989).

	D'aprs : "Apprentissage d'une chane de Markov cache.
	Problmes numriques lis  l'application  l'image." 
	de T.Brouard, M. Slimane, J.-P. Asselin de Beauville, 
	G. Venturini.

	Version / 27 juillet 2010.

*/

// -------------------------------------------------------------------------------------------------------------

#include "ext.h"							
#include "ext_obex.h"
#include "stdlib.h"
#include "math.h"	
#include "time.h"
#include "ext_common.h"		
#include "ext_systhread.h"	

// -------------------------------------------------------------------------------------------------------------
	
#define	TAILLE_GRAF_MAX		50
#define	TAILLE_LIST_MAX		256
#define TAILLE_LCD_MIN		150
#define	TAILLE_LCD_MAX		400
#define	LCD_COLOR_MULT		5
#define	LCD_LARGEUR_MULT	3
#define ERROR_INF			1000

// -------------------------------------------------------------------------------------------------------------

typedef struct _noeud {
	double		*transition ;
	double		*distribution ;
	double		*alpha ;
	double		*beta ;
	double		*temp_transition ;
	double		*temp_distribution ;
	double		*temp_alpha ;
	} t_noeud ;
	
typedef struct _foxtrot {
	t_object			ob ;
	long				taille_graf ;
	long				taille_alphabet ;
	long				taille_vecteur ;
	long				taille_lcd ;
	long				graf_pos ;
	long				flag ;
	long				lock ;
	float				apprentissage ;
	long				*vecteur ;
	double				*initialisation ;
	double				*temp_initialisation ;
	double				*coef ;
	t_noeud				*graf ;
	void				*m_qelem ;
	void				*m_clock ;
	void				*m_out1 ;
	void				*m_out2 ;
	void				*m_out3 ;
	void				*m_out4 ;
	t_systhread_mutex   c_mutex ;
	} t_foxtrot ;

enum {	
	DEF_TAILLE_GRAF = 6,
	DEF_APPRENTISSAGE = 1,
	DEF_TAILLE_ALPHABET = 128,
	DEF_TAILLE_LCD = 150,
	} ;
	
// -------------------------------------------------------------------------------------------------------------	
	
void	*foxtrot_new			(t_symbol *s, long argc, t_atom *argv) ;
void	foxtrot_free			(t_foxtrot *x) ;
void	foxtrot_assist			(t_foxtrot *x, void *b, long m, long a, char *s) ;

void	foxtrot_dumpmax			(t_foxtrot *x) ;
void	foxtrot_task_max		(t_foxtrot *x) ;
void	foxtrot_dumplcd			(t_foxtrot *x) ;
void	foxtrot_qtask_lcd		(t_foxtrot *x) ;
void	foxtrot_bang			(t_foxtrot *x) ;
void	foxtrot_clear			(t_foxtrot *x) ;
void	foxtrot_int				(t_foxtrot *x, long n) ;
void	foxtrot_list			(t_foxtrot *x, t_symbol *s, long argc, t_atom *argv) ;
void	foxtrot_vraisemblance	(t_foxtrot *x, t_symbol *s, long argc, t_atom *argv) ;

void	graf_init				(t_foxtrot *x) ;
void	graf_init_vecteur		(t_foxtrot *x, long argc, t_atom *argv) ;
void	graf_init_step			(t_foxtrot *x) ;
void	graf_free				(t_foxtrot *x) ;
void	graf_clear				(t_foxtrot *x) ;
long	graf_step				(t_foxtrot *x) ;

void	calculer_forward		(t_foxtrot *x) ;
void	calculer_backward		(t_foxtrot *x) ;
void	calculer_estimation		(t_foxtrot *x) ;
void	calculer_apprentissage	(t_foxtrot *x) ;

double	calculer_vraisemblance	(t_foxtrot *x) ;

double	calculer_coefficient			(t_foxtrot *x, long t) ;
double	calculer_produit_coefficient	(t_foxtrot *x) ;
double	calculer_gamma					(t_foxtrot *x, long n, long t) ;
double	calculer_sigma					(t_foxtrot *x, long a, long b) ;

void	stochastic_vecteur				(double *ptr, long n) ;

// -------------------------------------------------------------------------------------------------------------

static t_class *foxtrot_class ;

int main (void)
	{	
		t_class *c = NULL ;
		
		c = class_new ("foxtrot", (method)foxtrot_new, (method)foxtrot_free, (long)sizeof(t_foxtrot), 0L, A_GIMME, 0) ;

		class_addmethod (c, (method)foxtrot_assist,			"assist", A_CANT, 0) ; 
		class_addmethod (c, (method)foxtrot_bang,			"bang", 0) ;
		class_addmethod (c, (method)foxtrot_dumpmax,		"dump_max", 0) ;
		class_addmethod (c, (method)foxtrot_dumplcd,		"dump_lcd", 0) ;
		class_addmethod (c, (method)foxtrot_list,			"list", A_GIMME, 0) ;
		class_addmethod (c, (method)foxtrot_vraisemblance,	"vraisemblance", A_GIMME, 0) ;
		class_addmethod (c, (method)foxtrot_clear,			"clear", 0) ;
		class_addmethod (c, (method)foxtrot_int,			"int", A_LONG, 0) ;
		
		CLASS_METHOD_ATTR_PARSE (c, "int", "undocumented", gensym ("long"), ATTR_FLAGS_NONE, "1") ;
		
		CLASS_ATTR_LONG			(c, "lcd",				ATTR_SET_DEFER, t_foxtrot, taille_lcd) ;	
		CLASS_ATTR_FLOAT		(c, "apprentissage",	ATTR_FLAGS_NONE, t_foxtrot, apprentissage) ;
		
		CLASS_ATTR_LABEL		(c, "lcd",				ATTR_FLAGS_NONE, "Lcd") ;
		CLASS_ATTR_LABEL		(c, "apprentissage",	ATTR_FLAGS_NONE, "Apprentissage Persistance") ;
		
		CLASS_ATTR_SAVE			(c, "lcd",				ATTR_FLAGS_NONE) ;
		CLASS_ATTR_SAVE			(c, "apprentissage",	ATTR_FLAGS_NONE) ;
		
		CLASS_ATTR_FILTER_CLIP	(c, "lcd",				TAILLE_LCD_MIN, TAILLE_LCD_MAX) ;  
		CLASS_ATTR_FILTER_CLIP	(c, "apprentissage",	0., FLT_MAX) ;
		
		CLASS_ATTR_ORDER		(c, "apprentissage",	ATTR_FLAGS_NONE, "1") ;
		CLASS_ATTR_ORDER		(c, "lcd",				ATTR_FLAGS_NONE, "2") ;
		
		class_register (CLASS_BOX, c) ; 
		
		foxtrot_class = c ;
		
		return MAX_ERR_NONE ;
	}

void *foxtrot_new (t_symbol *s, long argc, t_atom *argv)
	{
		t_foxtrot *x = NULL;
		
		if (x = (t_foxtrot *)object_alloc (foxtrot_class)) 
			{	
				srand (time(NULL)) ;
				
				x->taille_graf			= DEF_TAILLE_GRAF ;
				x->taille_alphabet		= DEF_TAILLE_ALPHABET ;
				x->taille_lcd			= DEF_TAILLE_LCD ;
				x->apprentissage		= DEF_APPRENTISSAGE ;
				x->flag					= 0 ;
				x->lock					= 0 ;

				if (atom_gettype (argv) == A_LONG)
					{
						long k = atom_getlong (argv) ;
						
						if (k > 1 && k <= TAILLE_GRAF_MAX)
							{
								x->taille_graf = k ;
							}
					}
	
				x->m_qelem	= qelem_new ((t_object *)x, (method)foxtrot_qtask_lcd) ;
				x->m_clock	= clock_new ((t_object *)x, (method)foxtrot_task_max) ;
				
				x->m_out4	= floatout ((t_object *)x) ;
				x->m_out3	= outlet_new ((t_object *)x, NULL) ;
				x->m_out2	= bangout ((t_object *)x) ;
				x->m_out1	= intout ((t_object *)x) ;	
				
				systhread_mutex_new (&x->c_mutex, SYSTHREAD_MUTEX_NORMAL) ;
					
				attr_args_process (x, argc, argv) ;
				
				graf_init (x) ;
			}
				
		return x ;
	}
	
void foxtrot_free (t_foxtrot *x)
	{ 
		graf_free (x) ;
		
		systhread_mutex_free (x->c_mutex) ;
		
		qelem_unset (x->m_qelem) ;
		qelem_free	(x->m_qelem) ;
		
		clock_unset (x->m_clock) ;
		object_free (x->m_clock) ;
	}

void foxtrot_assist (t_foxtrot *x, void *b, long m, long a, char *s)
	{
		if (m == ASSIST_INLET) 
			{ 
				sprintf (s, "(bang) (list) clear vraisemblance") ;
			} 
		else 
			{	
				switch (a) {
					case 0 :
						sprintf (s, "Navigate") ;
						break ;
					case 1:
						sprintf (s, "Bang if empty") ;
						break ;
					case 2 :
						sprintf (s, "Dump to lcd") ;
						break ;	
					case 3 :
						sprintf (s, "Ln vraisemblance") ;
						break ;
				}
			}
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void foxtrot_dumpmax (t_foxtrot *x) 
	{
		if (!x->lock)
			{
				x->lock = 1 ;
				
				if (x->graf)
					{
						long i ;
						
						post ("Taille MMC : %ld", x->taille_graf) ;
						post ("Taille vecteur : %ld", x->taille_vecteur) ;
						post ("Apprentissage : %f", x->apprentissage) ;
						
						if (x->initialisation)
							{
								long j ;
								
								for (j = 0 ; j < x->taille_graf ; j++)
									{
										post ("Initialisation %ld : %f", j, *(x->initialisation + j)) ;
									}
							}
					
						for (i = 0 ; i < MIN (x->taille_graf, 10) ; i++)
							{
								t_noeud	*nptr = x->graf + i ;
								
								if (nptr->transition)
									{
										long j ;
										
										for (j = 0 ; j < x->taille_graf ; j++)
											{
												post ("Transition %ld %ld : %f", i, j, *(nptr->transition + j)) ;
											}
									}
							}
						
						for (i = 0 ; i < MIN (x->taille_graf, 10) ; i++)
							{
								t_noeud	*nptr = x->graf + i ;
								
								if (nptr->distribution)
									{
										long j ;
										
										for (j = 0 ; j < x->taille_alphabet ; j++)
											{
												post ("Distribution %ld %ld : %f", i, j, *(nptr->distribution + j)) ;
											}
									}
							}
					}
					
				clock_fdelay (x->m_clock, 1000.) ;
			}
	}
	
void foxtrot_task_max (t_foxtrot *x)
	{
		x->lock = 0 ;
	}
	
void foxtrot_dumplcd (t_foxtrot *x) 
	{
		qelem_set (x->m_qelem) ;
	}

void foxtrot_qtask_lcd (t_foxtrot *x) 
	{
		long	alpha = x->taille_lcd / x->taille_alphabet ;
		long	beta = x->taille_lcd / x->taille_graf ;
		t_atom	lcdsize[2] ;
		
		atom_setlong (lcdsize, alpha * x->taille_alphabet * LCD_LARGEUR_MULT) ;
		atom_setlong (lcdsize + 1, beta * x->taille_graf) ;
		
		outlet_anything (x->m_out3, gensym("clear"), 0, NULL) ;
		outlet_anything (x->m_out3, gensym("size"), 2, lcdsize) ;
		
		if (x->graf)
			{
				long	i ;
				t_atom	argv[5] ;
				t_noeud *nptr = NULL ;
				
				for (i = 0 ; i < x->taille_graf ; i++)
					{
						nptr = x->graf + i ;
						
						if (nptr->distribution)
							{
								long j ;
								
								for (j = 0 ; j < x->taille_alphabet ; j++)
									{
										long	k = 0 ;
										float	val = (*(nptr->distribution + j)) * 10. * LCD_COLOR_MULT ;
										
										if (val > 1.5)
											{
												k = MIN (round (val) + 244, 255) ;
											}
										else if (val > 0.1)
											{
												k = 245 ;
											}
										
										atom_setlong (&argv[0], j * alpha * LCD_LARGEUR_MULT) ;
										atom_setlong (&argv[1], i * beta) ;
										atom_setlong (&argv[2], (j + 1) * alpha * LCD_LARGEUR_MULT) ;
										atom_setlong (&argv[3], (i + 1) * beta) ;
										atom_setlong (&argv[4], k) ;
										
										outlet_anything (x->m_out3, gensym ("paintrect"), 5, argv) ;
									}
							}
					}
			}
	}
	
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void foxtrot_bang (t_foxtrot *x)
	{
		if (!x->flag)
			{
				outlet_bang (x->m_out2) ;
			}
		else
			{
				long k = -1 ;
				
				if (x->graf_pos < 0)
					{
						graf_init_step (x) ;
					}
				
				k = graf_step (x) ;
					
				outlet_int (x->m_out1, k) ;
			}
	}
	
void foxtrot_clear (t_foxtrot *x)
	{
		graf_clear (x) ;
	}
	
void foxtrot_int (t_foxtrot *x, long n)
	{
		;
	}
	
void foxtrot_list (t_foxtrot *x, t_symbol *s, long argc, t_atom *argv)
	{	
		if (argc <= TAILLE_LIST_MAX)
			{
				if (!systhread_mutex_trylock (x->c_mutex))
					{
						graf_init_vecteur (x, argc, argv) ;
						
						calculer_forward (x) ;
						calculer_backward (x) ;
					
						if (calculer_produit_coefficient (x))
							{
								calculer_estimation (x) ;
								
								systhread_mutex_unlock (x->c_mutex) ;
								
								calculer_apprentissage (x) ;
								
								x->flag = 1 ;
							}
						else
							{
								systhread_mutex_unlock (x->c_mutex) ;
							}
					}
			}
	}
	
void foxtrot_vraisemblance (t_foxtrot *x, t_symbol *s, long argc, t_atom *argv)
	{	
		double k = 0. ;
		
		if (!systhread_mutex_trylock (x->c_mutex))
			{
				graf_init_vecteur (x, argc, argv) ;
				
				calculer_forward (x) ;
				k = calculer_vraisemblance (x) ;
				
				systhread_mutex_unlock (x->c_mutex) ;
			}
			
		if (!k)
			{
				outlet_float (x->m_out4, ERROR_INF) ;
			}
		else
			{
				outlet_float (x->m_out4, -log(k)) ;
			}
		
	}
	
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void graf_init (t_foxtrot *x)
	{
		x->graf					= NULL ;
		x->initialisation		= NULL ;
		x->temp_initialisation	= NULL ;
		x->vecteur				= NULL ;
		x->coef					= NULL ;
		x->taille_vecteur		= 0 ;
		x->graf_pos				= -1 ;
		
		if (x->graf = (t_noeud *)sysmem_newptr (sizeof(t_noeud) * x->taille_graf))
			{
				long	i ;
				t_noeud *nptr = NULL ;
				
				x->temp_initialisation = (double *)sysmem_newptr (sizeof(double) * x->taille_graf) ;
				
				if (x->initialisation = (double *)sysmem_newptr (sizeof(double) * x->taille_graf))  
					{ stochastic_vecteur (x->initialisation, x->taille_graf) ; }
				
				for (i = 0 ; i < x->taille_graf ; i++)
					{
						nptr = x->graf + i ;
						
						nptr->temp_transition = (double *)sysmem_newptr (sizeof(double) * x->taille_graf) ;
						nptr->temp_distribution = (double *)sysmem_newptr (sizeof(double) * x->taille_alphabet) ;
						
						if (nptr->transition = (double *)sysmem_newptr (sizeof(double) * x->taille_graf))  
							{ stochastic_vecteur (nptr->transition, x->taille_graf) ; }

						if (nptr->distribution = (double *)sysmem_newptr (sizeof(double) * x->taille_alphabet))  
							{ stochastic_vecteur (nptr->distribution, x->taille_alphabet) ; }
							
						nptr->alpha = NULL ;
						nptr->beta	= NULL ;
						nptr->temp_alpha = NULL ;
					}				
			}
	}

// -------------------------------------------------------------------------------------------------------------

void graf_init_vecteur (t_foxtrot *x, long argc, t_atom *argv)   
	{
		if (!x->vecteur)
			{ 
				x->vecteur = (long *)sysmem_newptr (sizeof(long) * argc) ; 
			}
		else
			{
				if (x->taille_vecteur != argc)
					{ 
						x->vecteur = (long *)sysmem_resizeptr (x->vecteur, sizeof(long) * argc) ;
					}
			}
		
		if (x->taille_vecteur != argc)
			{
				long	j ;
				t_noeud *nptr ;
				
				if (x->coef)	
					{ x->coef = (double *)sysmem_resizeptr (x->coef, sizeof(double) * argc) ; }
				else
					{ x->coef = (double *)sysmem_newptr (sizeof(double) * argc) ; }
				
				for (j = 0 ; j < x->taille_graf ; j++)
					{
						nptr = x->graf + j ;
						
						if (nptr->alpha)	
							{ nptr->alpha = (double *)sysmem_resizeptr (nptr->alpha, sizeof(double) * argc) ; }
						else
							{ nptr->alpha = (double *)sysmem_newptr (sizeof(double) * argc) ; }
				
						if (nptr->beta)	
							{ nptr->beta = (double *)sysmem_resizeptr (nptr->beta, sizeof(double) * argc) ; }
						else
							{ nptr->beta = (double *)sysmem_newptr (sizeof(double) * argc) ; }
							
						if (nptr->temp_alpha)	
							{ nptr->temp_alpha = (double *)sysmem_resizeptr (nptr->temp_alpha, sizeof(double) * argc) ; }
						else
							{ nptr->temp_alpha = (double *)sysmem_newptr (sizeof(double) * argc) ; }
					}
												
				x->taille_vecteur = argc ;
			}

		if (x->vecteur)
			{
				long i ;
				
				for (i = 0 ; i < x->taille_vecteur ; i++)
					{
						if (atom_gettype (argv + i) != A_LONG)
							{
								*(x->vecteur + i) = rand () % 128 ;
							}
						else
							{
								*(x->vecteur + i) = CLIP (atom_getlong (argv + i), 0, 127) ;
							}
					}
			}
	}

// -------------------------------------------------------------------------------------------------------------

void graf_init_step (t_foxtrot *x)
	{
		long k = -1 ;
		
		if (x->graf && x->initialisation)
			{
				long	i ;
				double	w = 0. ;
				double	hasard = (double)rand () / (double)RAND_MAX ;
				
				for (i = 0 ; i < x->taille_graf ; i++)
					{
						w += *(x->initialisation + i) ;
						
						if (w > hasard)
							{
								k = i ;
								break ;
							}
					}
			}
		
		x->graf_pos = k ;
	}
	
// -------------------------------------------------------------------------------------------------------------
	
void graf_free (t_foxtrot *x)
	{
		if (x->initialisation)
			{ sysmem_freeptr (x->initialisation) ; }
		if (x->temp_initialisation)
			{ sysmem_freeptr (x->temp_initialisation) ; }
		if (x->vecteur)
			{ sysmem_freeptr (x->vecteur) ; }
		if (x->coef)
			{ sysmem_freeptr (x->coef) ; }
			
		if (x->graf)
			{
				long	i ;
				t_noeud *nptr = NULL ;
				
				for (i = 0 ; i < x->taille_graf ; i++)
					{
						nptr = x->graf + i ;
						
						if (nptr->transition)
							{ sysmem_freeptr (nptr->transition) ; }
						if (nptr->distribution)
							{ sysmem_freeptr (nptr->distribution) ; }
						if (nptr->alpha)
							{ sysmem_freeptr (nptr->alpha) ; }
						if (nptr->beta)
							{ sysmem_freeptr (nptr->beta) ; }
						if (nptr->temp_alpha)
							{ sysmem_freeptr (nptr->temp_alpha) ; }
						if (nptr->temp_transition)
							{ sysmem_freeptr (nptr->temp_transition) ; }
						if (nptr->temp_distribution)
							{ sysmem_freeptr (nptr->temp_distribution) ; }
					}
				
				sysmem_freeptr (x->graf) ;
			}
	}

// -------------------------------------------------------------------------------------------------------------

void graf_clear (t_foxtrot *x)
	{
		x->graf_pos	= -1 ;
		x->flag = 0 ;
	
		if (x->initialisation)  
			{ stochastic_vecteur (x->initialisation, x->taille_graf) ; }
		
		if (x->graf)
			{
				long	i ;
				t_noeud *nptr = NULL ;
		
				for (i = 0 ; i < x->taille_graf ; i++)
					{
						nptr = x->graf + i ;
						
						if (nptr->transition)  
							{ stochastic_vecteur (nptr->transition, x->taille_graf) ; }

						if (nptr->distribution)  
							{ stochastic_vecteur (nptr->distribution, x->taille_alphabet) ; }
					}
			}		
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

long graf_step (t_foxtrot *x)
	{
		long k = -1 ;
		
		if (x->graf && x->graf_pos > -1)
			{
				t_noeud *nptr = x->graf + (x->graf_pos) ;
				
				if (nptr->distribution)
					{
						long	i ;
						double	w = 0. ;
						double	hasard = (double)rand () / (double)RAND_MAX ;
		
						for (i = 0 ; i < x->taille_alphabet ; i++)
							{
								w += *(nptr->distribution + i) ;
								
								if (w > hasard)
									{
										k = i ;
										break ;
									}
							}
					}
				
				if (nptr->transition)
					{
						long	i ;
						double	w = 0. ;
						double	hasard = (double)rand () / (double)RAND_MAX ;
		
						for (i = 0 ; i < x->taille_graf ; i++)
							{
								w += *(nptr->transition + i) ;
								
								if (w > hasard)
									{
										x->graf_pos = i ;
										break ;
									}
							}
					}
			}
					
		return k ;
	}
	
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void calculer_forward (t_foxtrot *x)
	{
		if (x->graf && x->initialisation && x->vecteur)
			{
				long	n, t ;
				double	coef_zero ;
				
				for (n = 0 ; n < x->taille_graf ; n++)
					{
						t_noeud	*nptr = x->graf + n ;
						
						if (nptr->distribution && nptr->temp_alpha)
							{
								*(nptr->temp_alpha) = *(x->initialisation + n) * (*(nptr->distribution + *(x->vecteur))) ; 
							}
					}
					
				coef_zero = calculer_coefficient (x, 0) ;
				
				for (n = 0 ; n < x->taille_graf ; n++)
					{
						t_noeud	*nptr = x->graf + n ;
						
						if (nptr->alpha && nptr->temp_alpha)
							{
								*(nptr->alpha) = *(nptr->temp_alpha) * coef_zero ;
							}
					}
				
				for (t = 0 ; t < (x->taille_vecteur - 1) ; t++)  
					{
						long	j ;
						double	coef_temps ;
						
						for (j = 0 ; j < x->taille_graf ; j ++)
							{
								t_noeud *nptr1 = x->graf + j ;
								
								if (nptr1->distribution && nptr1->transition && nptr1->temp_alpha)
									{
										long	i ;
										double	k = 0. ;
										
										for (i = 0 ; i < x->taille_graf ; i++)
											{
												t_noeud *nptr2 = x->graf + i ;
												
												if (nptr2->alpha && nptr2->transition)
													{
														k += *(nptr2->alpha + t) * (*(nptr2->transition + j)) ;
													}
											}
											
										*(nptr1->temp_alpha + (t + 1)) = *(nptr1->distribution + *(x->vecteur + (t + 1))) * k ;
									}	
							}
							
						coef_temps = calculer_coefficient (x, t + 1) ;
						
						for (j = 0 ; j < x->taille_graf ; j++)
							{
								t_noeud *nptr = x->graf + j ;
								
								if (nptr->alpha && nptr->temp_alpha)
									{
										*(nptr->alpha + (t + 1)) = *(nptr->temp_alpha + (t + 1)) * coef_temps ;
									}
							}
					}
			}
	}

void calculer_backward (t_foxtrot *x)
	{
		if (x->graf && x->vecteur && x->coef)
			{
				long n, t ;
				
				for (n = 0 ; n < x->taille_graf ; n++)
					{
						t_noeud	*nptr = x->graf + n ;
						
						if (nptr->beta)
							{
								*(nptr->beta + (x->taille_vecteur - 1)) = *(x->coef + (x->taille_vecteur - 1)) ; 
							}
					}
				
				for (t = (x->taille_vecteur - 2) ; t > -1 ; t--)
					{
						long	i ;
						double	coef_temps = *(x->coef + t) ;
						
						for (i = 0 ; i < x->taille_graf ; i ++)
							{
								t_noeud *nptr1 = x->graf + i ;
								
								if (nptr1->transition && nptr1->beta)
									{
										long	j ;
										double	k = 0. ;
										
										for (j = 0 ; j < x->taille_graf ; j++)
											{
												t_noeud *nptr2 = x->graf + j ;

												if (nptr2->beta && nptr2->distribution)
													{
														k += *(nptr1->transition + j) * (*(nptr2->beta + (t + 1))) * (*(nptr2->distribution + *(x->vecteur + (t + 1)))) ;
													}
											}
										
										*(nptr1->beta + t) = coef_temps * k ;
									}	
							}
					}
			}
	}

// -------------------------------------------------------------------------------------------------------------

void calculer_estimation (t_foxtrot *x) 
	{
		if (x->graf && x->vecteur && x->coef)
			{
				long i ;
				
				if (x->temp_initialisation) 
					{
						long j ;
						
						for (j = 0 ; j < x->taille_graf ; j++)
							{
								*(x->temp_initialisation + j) = calculer_gamma (x, j, 0) ;
							}
					}
				
				for (i = 0 ; i < x->taille_graf ; i++)
					{
						t_noeud *nptr = x->graf + i ;
						
						if (nptr->temp_transition)
							{
								long	j ;
								double	k = 0. ;
								double  memo[TAILLE_GRAF_MAX] ;
								
								for (j = 0 ; j < x->taille_graf ; j++)
									{
										memo[j] = calculer_sigma (x, i, j) ;
										k += memo[j] ;
									}
								
								for (j = 0 ; j < x->taille_graf ; j++)
									{
										if (k)
											{
												*(nptr->temp_transition + j) = memo[j] / k ;
											}
									}
							}
						
						if (nptr->temp_distribution)
							{
								long	j, t ;
								double	w = 0. ;
								double	memo[TAILLE_LIST_MAX] ;
								
								for (t = 0 ; t < x->taille_vecteur ; t++)
									{
										memo[t]	= calculer_gamma (x, i, t) ;
										w += memo[t] ;
									}
								
								if (w)
									{
										for (j = 0 ; j < x->taille_alphabet ; j++)
											{
												double k = 0. ;
												
												for (t = 0 ; t < x->taille_vecteur ; t++)
													{
														if (*(x->vecteur + t) == j)
															{
																k += memo[t] ;	
															}
													}
											
												*(nptr->temp_distribution + j) = k / w ;
											}
									}
							}
					}
			}
	}

// -------------------------------------------------------------------------------------------------------------

void calculer_apprentissage (t_foxtrot *x)
	{
		if (x->graf)
			{
				long	i ;
				t_noeud *nptr ;
				
				if (x->initialisation && x->temp_initialisation) 
					{
						long j ;
				
						for (j = 0 ; j < x->taille_graf ; j++)
							{
								*(x->initialisation + j) = (x->apprentissage * (*(x->initialisation + j)) + (*(x->temp_initialisation + j))) / (x->apprentissage + 1.) ;
							}
					}
				
				for (i = 0 ; i < x->taille_graf ; i++)
					{
						nptr = x->graf + i ;
						
						if (nptr->transition && nptr->temp_transition)
							{
								long j ;
								
								for (j = 0 ; j < x->taille_graf ; j++)
									{
										*(nptr->transition + j) = (x->apprentissage * (*(nptr->transition + j)) + (*(nptr->temp_transition + j))) / (x->apprentissage + 1.) ;
									}
							}
						
						if (nptr->distribution && nptr->temp_distribution)
							{
								long j ;
								
								for (j = 0 ; j < x->taille_alphabet ; j++)
									{
										*(nptr->distribution + j) = (x->apprentissage * (*(nptr->distribution + j)) + (*(nptr->temp_distribution + j))) / (x->apprentissage + 1.) ;	
									}
							}
					}
			}
	}

// -------------------------------------------------------------------------------------------------------------

double calculer_vraisemblance (t_foxtrot *x) 
	{
		double k = 0. ;
		
		if (x->coef)
			{
				long i ;
				
				for (i = 0 ; i < x->taille_vecteur ; i++)
					{
						k += *(x->coef + i) ;
					}
			}
			
		if (k)
			{
				k = 1 / k ;
			}
			
		return k ;
	}
	
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

double calculer_coefficient (t_foxtrot *x, long t) 
	{
		double	k = 0. ;
		
		if (x->coef && x->graf)
			{
				long	i ;
				double	w = 0. ;
				t_noeud *nptr ;
				
				for (i = 0 ; i < x->taille_graf ; i++)
					{
						nptr = x->graf + i ;
						
						if (nptr->temp_alpha) 
							{
								w += *(nptr->temp_alpha + t) ;
							}
					}
				
				if (w) 
					{	
						k = 1. / w ;
						*(x->coef + t) = k ;
					}
				else
					{
						*(x->coef + t) = 0. ;  
					}
			}
		
		return k ;
	}

// -------------------------------------------------------------------------------------------------------------

double calculer_produit_coefficient (t_foxtrot *x) 
	{
		double k = 0. ;
		
		if (x->coef)
			{
				long i ;
				
				k = *(x->coef) ;
				
				for (i = 1 ; i < x->taille_vecteur ; i++)
					{
						k *= (*(x->coef + i)) ;
					}
			}
		
		return k ;
	}

// -------------------------------------------------------------------------------------------------------------

double calculer_gamma (t_foxtrot *x, long n, long t)
	{
		double	k = 0. ;
		t_noeud *nptr = x->graf + n ;
				
		if (nptr->beta && nptr->alpha)
			{
				k = (*(nptr->alpha + t) * (*(nptr->beta + t))) / *(x->coef + t) ;	
			}
				
		return k ;
	}

double calculer_sigma (t_foxtrot *x, long a, long b)
	{
		double	k = 0. ;
		t_noeud	*nptr1 = x->graf + a ;
		t_noeud	*nptr2 = x->graf + b ;
		
		if (nptr1->alpha && nptr1->transition && nptr2->distribution && nptr2->beta)
			{ 
				long t ;
			
				for (t = 0 ; t < (x->taille_vecteur - 1) ; t++)
					{	
						k += *(nptr1->alpha + t) * (*(nptr1->transition + b)) * (*(nptr2->distribution + *(x->vecteur + (t + 1)))) * (*(nptr2->beta + (t + 1))) ;
					}
			}
			
		return k ;
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void stochastic_vecteur (double *ptr, long n)
	{
		long	i ;
		double	f = (rand () % 6) + 1 ;
		double	s = 0. ;
		
		for (i = 0 ; i < n ; i++)
			{
				double k = pow ((double)rand () / (double)RAND_MAX, f) ;
				
				*(ptr + i) = k ;
				s += k ;
			}
		
		for (i = 0 ; i < n ; i++)
			{
				*(ptr + i) = (*(ptr + i)) / s ;
			}
	}
	
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------:x